@@ -3,7 +3,8 @@ use darling::FromMeta;
 use proc_macro2::{Ident, TokenStream};
 use quote::{quote, ToTokens};
 use std::collections::HashMap;
-use syn::{AttributeArgs, Expr, GenericParam, ItemEnum, ItemStruct, ItemType, Lit, Type, Visibility};
+use syn::spanned::Spanned;
+use syn::{AttributeArgs, Expr, Field, GenericParam, ItemEnum, ItemStruct, ItemType, Lit, Type, Visibility};
 
 #[derive(Debug, FromMeta, Clone)]
 pub struct FFITypeAttributes {
@@ -19,6 +20,9 @@ pub struct FFITypeAttributes {
     #[darling(default)]
     skip: HashMap<String, ()>,
 
+    #[darling(default)]
+    visibility: HashMap<String, String>,
+
     #[darling(default)]
     name: Option<String>,
 
@@ -27,6 +31,42 @@ pub struct FFITypeAttributes {
 
     #[darling(default)]
     debug: bool,
+
+    #[darling(default, rename = "unsafe")]
+    unsfe: bool,
+}
+
+impl FFITypeAttributes {
+    pub fn assert_valid(&self) {
+        if (!self.surrogates.is_empty() || !self.skip.is_empty()) && !self.unsfe {
+            panic!("When using `surrogate` or `skip` you must also specify `unsafe`.")
+        }
+    }
+
+    pub fn visibility_for_field(&self, field: &Field, name: &str) -> TokenStream {
+        let mut rval = match &field.vis {
+            Visibility::Public(_) => quote! { interoptopus::lang::c::Visibility::Public },
+            _ => quote! { interoptopus::lang::c::Visibility::Private },
+        };
+
+        if let Some(x) = self.visibility.get(name) {
+            rval = match x.as_str() {
+                "public" => quote! { interoptopus::lang::c::Visibility::Public },
+                "private" => quote! { interoptopus::lang::c::Visibility::Private },
+                _ => panic!("Visibility must be `public` or `private`"),
+            };
+        }
+
+        if let Some(x) = self.visibility.get("_") {
+            rval = match x.as_str() {
+                "public" => quote! { interoptopus::lang::c::Visibility::Public },
+                "private" => quote! { interoptopus::lang::c::Visibility::Private },
+                _ => panic!("Visibility must be `public` or `private`"),
+            };
+        }
+
+        rval
+    }
 }
 
 fn derive_variant_info(item: ItemEnum, idents: &[Ident], names: &[String], values: &[i32], docs: &[String]) -> TokenStream {
@@ -51,6 +91,10 @@ fn derive_variant_info(item: ItemEnum, idents: &[Ident], names: &[String], value
 }
 
 pub fn ffi_type_enum(attr: &FFITypeAttributes, input: TokenStream, item: ItemEnum) -> TokenStream {
+    let doc_line = extract_doc_lines(&item.attrs).join("\n");
+
+    attr.assert_valid();
+
     let span = item.ident.span();
     let name = item.ident.to_string();
     let name_ident = syn::Ident::new(&name, span);
@@ -59,11 +103,8 @@ pub fn ffi_type_enum(attr: &FFITypeAttributes, input: TokenStream, item: ItemEnu
     let mut variant_idents = Vec::new();
     let mut variant_values = Vec::new();
     let mut variant_docs = Vec::new();
-
     let mut next_id = 0;
 
-    let doc_line = extract_doc_lines(&item.attrs).join("\n");
-
     for variant in &item.variants {
         let ident = variant.ident.to_string();
         let variant_doc_line = extract_doc_lines(&variant.attrs).join("\n");
@@ -132,41 +173,108 @@ pub fn ffi_type_enum(attr: &FFITypeAttributes, input: TokenStream, item: ItemEnu
     }
 }
 
+// Various Struct examples
+//
+// ```
+// pub struct S1<'a, T>
+// where
+//     T: 'static,
+//     T: CTypeInfo,
+// {
+//     pub x: u32,
+//     pub p: PhantomData<&'a T>,
+// }
+//
+// pub struct S2<'a, T: Clone>
+// {
+//     pub x: u32,
+//     pub p: PhantomData<&'a T>,
+// }
+//
+// pub struct Tupled(pub u8);
+//
+// pub struct S3<const N: usize>
+// {
+//     pub f: [u32; N],
+// }
+//
+// struct Weird2<'a, T: Clone, const N: usize>
+//     where
+//         T: Copy + Copy + 'a,
+//         T: Debug,
+// {
+//     t: &'a T,
+//     a: [T; N],
+//     r: &'a u8,
+// }
+//
+// ```
+//
 pub fn ffi_type_struct(attr: &FFITypeAttributes, input: TokenStream, item: ItemStruct) -> TokenStream {
-    let name = item.ident.to_string();
-    let name_ident = syn::Ident::new(&name, item.ident.span());
-    let c_struct_name = attr.name.clone().unwrap_or(name);
     let namespace = attr.namespace.clone().unwrap_or_else(|| "".to_string());
+    let doc_line = extract_doc_lines(&item.attrs).join("\n");
+
+    attr.assert_valid();
+
+    let struct_ident_str = item.ident.to_string();
+    let struct_ident = syn::Ident::new(&struct_ident_str, item.ident.span());
+    let struct_ident_c = attr.name.clone().unwrap_or(struct_ident_str);
 
-    let mut generic_params = Vec::new();
     let mut field_names = Vec::new();
     let mut field_types = Vec::new();
     let mut field_docs = Vec::new();
-    let mut field_visibility = Vec::new();
-    let mut generic_params_needing_bounds = Vec::new();
+    let mut field_visibilities = Vec::new();
 
-    let mut generics_params = quote! {};
-    let mut where_clause = quote! {};
-
-    let doc_line = extract_doc_lines(&item.attrs).join("\n");
+    let mut has_generics = false;
+    let mut generic_params_needing_ctypeinfo_bounds = Vec::new();
+    let mut generic_parameter_tokens = Vec::new();
+    let mut generic_struct_tokens = Vec::new();
+    let mut generic_where_tokens = Vec::new();
 
     for generic in &item.generics.params {
-        generic_params.push(quote! { #generic });
+        match generic {
+            GenericParam::Lifetime(lt) => {
+                let ident = lt.lifetime.ident.clone();
+                let lt = syn::Lifetime::new(&format!("'{}", ident.to_string()), item.span());
+                generic_parameter_tokens.push(quote! { #lt });
+                generic_struct_tokens.push(quote! { #lt });
+            }
+            GenericParam::Type(ty) => {
+                let ident = ty.ident.clone();
+                let whre = ty.bounds.to_token_stream();
+                generic_parameter_tokens.push(quote! { #ident });
+                generic_struct_tokens.push(quote! { #ident });
+                generic_params_needing_ctypeinfo_bounds.push(ty.ident.clone());
+                generic_where_tokens.push(quote! { #ident: interoptopus::lang::rust::CTypeInfo });
+                if !whre.to_string().is_empty() {
+                    generic_where_tokens.push(quote! { #ident: #whre });
+                }
+            }
+            GenericParam::Const(x) => {
+                let ident = x.ident.clone();
+                let ty = x.ty.to_token_stream();
+                generic_parameter_tokens.push(quote! { const #ident: #ty });
+                generic_struct_tokens.push(quote! { #ident });
+            }
+        }
 
-        if let GenericParam::Type(ty) = generic {
-            let ident = ty.ident.clone();
-            generic_params_needing_bounds.push(ident);
+        has_generics = true;
+    }
+
+    if let Some(whre) = item.generics.where_clause {
+        for pred in whre.predicates {
+            generic_where_tokens.push(quote! { #pred });
         }
     }
 
-    if !generic_params.is_empty() {
-        generics_params = quote! { < #(#generic_params),* > };
+    let mut param_param = quote! {};
+    let mut param_struct = quote! {};
+    let mut param_where = quote! {};
 
-        if !generic_params_needing_bounds.is_empty() {
-            if let Some(where_cl) = item.generics.where_clause {
-                where_clause = quote! { #where_cl }
-            }
-        }
+    if has_generics {
+        param_param = quote! { < #(#generic_parameter_tokens),* > };
+        param_struct = quote! { < #(#generic_struct_tokens),* > };
+        param_where = quote! { where #(#generic_where_tokens),*  };
     }
 
     for (i, field) in item.fields.iter().enumerate() {
@@ -176,14 +284,11 @@ pub fn ffi_type_struct(attr: &FFITypeAttributes, input: TokenStream, item: ItemS
             continue;
         }
 
-        let visibility = match &field.vis {
-            Visibility::Public(_) => quote! { interoptopus::lang::c::Visibility::Public },
-            _ => quote! { interoptopus::lang::c::Visibility::Private },
-        };
+        let visibility = attr.visibility_for_field(field, &name);
 
         field_names.push(name.clone());
         field_docs.push(extract_doc_lines(&field.attrs).join("\n"));
-        field_visibility.push(visibility);
+        field_visibilities.push(visibility);
 
         if attr.surrogates.contains_key(&name) {
             let lookup = attr.surrogates.get(&name).unwrap();
@@ -230,37 +335,37 @@ pub fn ffi_type_struct(attr: &FFITypeAttributes, input: TokenStream, item: ItemS
         }
     };
 
-    let fields = if attr.opaque {
+    let let_fields = if attr.opaque {
         quote! {}
     } else {
         quote! {
             #({
                 let documentation = interoptopus::lang::c::Documentation::from_line(#field_docs);
                 let the_type = #field_types;
-                let field = interoptopus::lang::c::Field::with_documentation(#field_names.to_string(), the_type, #field_visibility, documentation);
+                let field = interoptopus::lang::c::Field::with_documentation(#field_names.to_string(), the_type, #field_visibilities, documentation);
                 fields.push(field);
             })*
         }
     };
 
-    let name = if let Some(name) = &attr.name {
+    let let_name = if let Some(name) = &attr.name {
         quote! {
             let name = #name.to_string();
         }
     } else {
         quote! {
             #({
-                generics.push(<#generic_params_needing_bounds as interoptopus::lang::rust::CTypeInfo>::type_info().name_within_lib());
+                generics.push(<#generic_params_needing_ctypeinfo_bounds as interoptopus::lang::rust::CTypeInfo>::type_info().name_within_lib());
             })*
 
-            let name = format!("{}{}", #c_struct_name.to_string(), generics.join(""));
+            let name = format!("{}{}", #struct_ident_c.to_string(), generics.join(""));
         }
     };
 
     quote! {
         #input
 
-        unsafe impl #generics_params interoptopus::lang::rust::CTypeInfo for #name_ident #generics_params #where_clause {
+        unsafe impl #param_param interoptopus::lang::rust::CTypeInfo for #struct_ident #param_struct #param_where {
 
             fn type_info() -> interoptopus::lang::c::CType {
                 let documentation = interoptopus::lang::c::Documentation::from_line(#doc_line);
@@ -268,9 +373,9 @@ pub fn ffi_type_struct(attr: &FFITypeAttributes, input: TokenStream, item: ItemS
                 let mut fields: std::vec::Vec<interoptopus::lang::c::Field> = std::vec::Vec::new();
                 let mut generics: std::vec::Vec<String> = std::vec::Vec::new();
 
-                #name
+                #let_name
 
-                #fields
+                #let_fields
 
                 #rval_builder
             }
